"""Unit of work"""

from __future__ import annotations

import abc
from collections.abc import Generator
from types import TracebackType
from typing import Optional

from purgatory.domain.messages import Message
from purgatory.service._sync.repository import (
    SyncAbstractRepository,
    SyncInMemoryRepository,
    SyncRedisRepository,
)


class SyncAbstractUnitOfWork(abc.ABC):
    contexts: SyncAbstractRepository

    def collect_new_events(self) -> Generator[Message, None, None]:
        while self.contexts.messages:
            yield self.contexts.messages.pop(0)

    def initialize(self) -> None:  # noqa B027
        """Override to initialize  repositories."""

    def __enter__(self) -> SyncAbstractUnitOfWork:
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        """Rollback in case of exception."""
        if exc:
            self.rollback()

    @abc.abstractmethod
    def commit(self) -> None:
        """Commit the transation."""

    @abc.abstractmethod
    def rollback(self) -> None:
        """Rollback the transation."""


class SyncInMemoryUnitOfWork(SyncAbstractUnitOfWork):
    def __init__(self) -> None:
        self.contexts = SyncInMemoryRepository()

    def commit(self) -> None:
        """Do nothing."""

    def rollback(self) -> None:
        """Do nothing."""


class SyncRedisUnitOfWork(SyncAbstractUnitOfWork):
    def __init__(self, url: str) -> None:
        self.contexts = SyncRedisRepository(url)

    def initialize(self) -> None:
        self.contexts.initialize()

    def commit(self) -> None:
        """Do nothing."""

    def rollback(self) -> None:
        """Do nothing."""



"""
Propagate commands and events to every registered handles.

"""

import logging
from collections import defaultdict
from typing import Any, TypeVar, cast

from purgatory.domain.messages.base import Command, Event, Message

from ..typing import AsyncCommandHandler, AsyncEventHandler, AsyncMessageHandler
from .unit_of_work import AsyncAbstractUnitOfWork

log = logging.getLogger(__name__)

TCommand = TypeVar("TCommand", bound=Command)
TEvent = TypeVar("TEvent", bound=Event)


class ConfigurationError(RuntimeError):
    """Prevents bad usage of the add_listener."""


class AsyncMessageRegistry:
    """Store all the handlers for commands an events."""

    def __init__(self) -> None:
        self.commands_registry: dict[type[Command], AsyncCommandHandler[Command]] = {}
        self.events_registry: dict[type[Event], list[AsyncEventHandler[Event]]] = (
            defaultdict(list)
        )

    def add_listener(
        self, msg_type: type[Message], callback: AsyncMessageHandler[Any, Any]
    ) -> None:
        if issubclass(msg_type, Command):
            if msg_type in self.commands_registry:
                raise ConfigurationError(
                    f"{msg_type} command has been registered twice"
                )
            self.commands_registry[msg_type] = cast(
                AsyncCommandHandler[Command], callback
            )
        elif issubclass(msg_type, Event):
            self.events_registry[msg_type].append(
                cast(AsyncEventHandler[Event], callback)
            )
        else:
            raise ConfigurationError(
                f"Invalid usage of the listen decorator: "
                f"type {msg_type} should be a command or an event"
            )

    def remove_listener(
        self, msg_type: type, callback: AsyncMessageHandler[Any, Any]
    ) -> None:
        if issubclass(msg_type, Command):
            if msg_type not in self.commands_registry:
                raise ConfigurationError(f"{msg_type} command has not been registered")
            del self.commands_registry[msg_type]
        elif issubclass(msg_type, Event):
            try:
                self.events_registry[msg_type].remove(
                    cast(AsyncEventHandler[Event], callback)
                )
            except ValueError as exc:
                raise ConfigurationError(
                    f"{msg_type} event has not been registered"
                ) from exc
        else:
            raise ConfigurationError(
                f"Invalid usage of the listen decorator: "
                f"type {msg_type} should be a command or an event"
            )

    async def handle(self, message: Message, uow: AsyncAbstractUnitOfWork) -> Any:
        """
        Notify listener of that event registered with `messagebus.add_listener`.
        Return the first event from the command.
        """
        queue = [message]
        idx = 0
        ret = None
        while queue:
            message = queue.pop(0)
            if not isinstance(message, (Command, Event)):
                raise RuntimeError(f"{message} was not an Event or Command")
            msg_type = type(message)
            if msg_type in self.commands_registry:
                cmdret = await self.commands_registry[cast(type[Command], msg_type)](
                    cast(Command, message), uow
                )
                if idx == 0:
                    ret = cmdret
                queue.extend(uow.collect_new_events())
            elif msg_type in self.events_registry:
                for callback in self.events_registry[cast(type[Event], msg_type)]:
                    await callback(cast(Event, message), uow)
                    queue.extend(uow.collect_new_events())
            idx += 1
        return ret



from dataclasses import dataclass
from typing import Optional

from purgatory.typing import TTL, StateName, Threshold

from .base import Event


@dataclass(frozen=True)
class CircuitBreakerCreated(Event):
    name: str
    threshold: Threshold
    ttl: TTL


@dataclass(frozen=True)
class ContextChanged(Event):
    name: str
    state: StateName
    opened_at: Optional[float]


@dataclass(frozen=True)
class CircuitBreakerFailed(Event):
    name: str
    failure_count: int


@dataclass(frozen=True)
class CircuitBreakerRecovered(Event):
    name: str



from importlib import metadata

__version__ = metadata.version("purgatory")


from purgatory.domain.messages import Event
from purgatory.domain.messages.events import (
    CircuitBreakerCreated,
    CircuitBreakerFailed,
    CircuitBreakerRecovered,
    ContextChanged,
)
from purgatory.service._async.circuitbreaker import AsyncCircuitBreakerFactory
from purgatory.service._async.unit_of_work import (
    AsyncAbstractUnitOfWork,
    AsyncInMemoryUnitOfWork,
    AsyncRedisUnitOfWork,
)
from purgatory.service._sync.circuitbreaker import SyncCircuitBreakerFactory
from purgatory.service._sync.unit_of_work import (
    SyncAbstractUnitOfWork,
    SyncInMemoryUnitOfWork,
    SyncRedisUnitOfWork,
)

__all__ = [
    "AsyncAbstractUnitOfWork",
    "AsyncCircuitBreakerFactory",
    "AsyncInMemoryUnitOfWork",
    "AsyncRedisUnitOfWork",
    "CircuitBreakerCreated",
    "CircuitBreakerFailed",
    "CircuitBreakerRecovered",
    "ContextChanged",
    "Event",
    "SyncCircuitBreakerFactory",
    "SyncAbstractUnitOfWork",
    "SyncInMemoryUnitOfWork",
    "SyncRedisUnitOfWork",
]



from functools import wraps
from types import TracebackType
from typing import Any, Callable, Optional

from purgatory.domain.messages.commands import CreateCircuitBreaker
from purgatory.domain.messages.events import (
    CircuitBreakerCreated,
    CircuitBreakerFailed,
    CircuitBreakerRecovered,
    ContextChanged,
)
from purgatory.domain.model import Context, ExcludeType
from purgatory.service._async.message_handlers import (
    inc_circuit_breaker_failure,
    register_circuit_breaker,
    reset_failure,
    save_circuit_breaker_state,
)
from purgatory.service._async.messagebus import AsyncMessageRegistry
from purgatory.service._async.unit_of_work import (
    AsyncAbstractUnitOfWork,
    AsyncInMemoryUnitOfWork,
)
from purgatory.typing import TTL, CircuitName, Hook, Threshold


class AsyncCircuitBreaker:
    def __init__(
        self,
        context: Context,
        uow: AsyncAbstractUnitOfWork,
        messagebus: AsyncMessageRegistry,
    ) -> None:
        self.context = context
        self.uow = uow
        self.messagebus = messagebus

    async def __aenter__(self) -> "AsyncCircuitBreaker":
        self.context.__enter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        self.context.__exit__(exc_type, exc, tb)
        while self.context.messages:
            await self.messagebus.handle(
                self.context.messages.pop(0),
                self.uow,
            )


class PublicEvent:
    def __init__(self, messagebus: AsyncMessageRegistry, hook: Hook) -> None:
        messagebus.add_listener(CircuitBreakerCreated, self.cb_created)
        messagebus.add_listener(ContextChanged, self.cb_state_changed)
        messagebus.add_listener(CircuitBreakerFailed, self.cb_failed)
        messagebus.add_listener(CircuitBreakerRecovered, self.cb_recovered)
        self.hook = hook

    def remove_listeners(self, messagebus: AsyncMessageRegistry) -> None:
        messagebus.remove_listener(CircuitBreakerCreated, self.cb_created)
        messagebus.remove_listener(ContextChanged, self.cb_state_changed)
        messagebus.remove_listener(CircuitBreakerFailed, self.cb_failed)
        messagebus.remove_listener(CircuitBreakerRecovered, self.cb_recovered)

    async def cb_created(
        self, event: CircuitBreakerCreated, uow: AsyncAbstractUnitOfWork
    ) -> None:
        self.hook(event.name, "circuit_breaker_created", event)

    async def cb_state_changed(
        self, event: CircuitBreakerCreated, uow: AsyncAbstractUnitOfWork
    ) -> None:
        self.hook(event.name, "state_changed", event)

    async def cb_failed(
        self, event: CircuitBreakerCreated, uow: AsyncAbstractUnitOfWork
    ) -> None:
        self.hook(event.name, "failed", event)

    async def cb_recovered(
        self, event: CircuitBreakerCreated, uow: AsyncAbstractUnitOfWork
    ) -> None:
        self.hook(event.name, "recovered", event)


class AsyncCircuitBreakerFactory:
    def __init__(
        self,
        default_threshold: Threshold = 5,
        default_ttl: TTL = 30,
        exclude: Optional[ExcludeType] = None,
        uow: Optional[AsyncAbstractUnitOfWork] = None,
    ):
        self.default_threshold = default_threshold
        self.default_ttl = default_ttl
        self.global_exclude = exclude or []
        self.uow = uow or AsyncInMemoryUnitOfWork()
        self.messagebus = AsyncMessageRegistry()
        self.messagebus.add_listener(CreateCircuitBreaker, register_circuit_breaker)
        self.messagebus.add_listener(ContextChanged, save_circuit_breaker_state)
        self.messagebus.add_listener(CircuitBreakerFailed, inc_circuit_breaker_failure)
        self.messagebus.add_listener(CircuitBreakerRecovered, reset_failure)
        self.listeners: dict[Hook, PublicEvent] = {}

    async def initialize(self) -> None:
        await self.uow.initialize()

    def add_listener(self, listener: Hook) -> None:
        self.listeners[listener] = PublicEvent(self.messagebus, listener)

    def remove_listener(self, listener: Hook) -> None:
        try:
            self.listeners[listener].remove_listeners(self.messagebus)
            del self.listeners[listener]
        except KeyError as exc:
            raise RuntimeError(f"{listener} is not listening {self}") from exc

    async def get_breaker(
        self,
        circuit: CircuitName,
        threshold: Optional[Threshold] = None,
        ttl: Optional[TTL] = None,
        exclude: Optional[ExcludeType] = None,
    ) -> AsyncCircuitBreaker:
        async with self.uow as uow:
            brk = await uow.contexts.get(circuit)
        if brk is None:
            async with self.uow as uow:
                bkr_threshold = threshold or self.default_threshold
                bkr_ttl = ttl or self.default_ttl
                brk = await self.messagebus.handle(
                    CreateCircuitBreaker(circuit, bkr_threshold, bkr_ttl),
                    self.uow,
                )
        brk.exclude_list = (exclude or []) + self.global_exclude
        return AsyncCircuitBreaker(brk, self.uow, self.messagebus)

    def __call__(
        self,
        circuit: str,
        threshold: Optional[Threshold] = None,
        ttl: Optional[TTL] = None,
        exclude: Optional[ExcludeType] = None,
    ) -> Any:
        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            @wraps(func)
            async def inner_coro(*args: Any, **kwargs: Any) -> Any:
                brk = await self.get_breaker(circuit, threshold, ttl, exclude)
                async with brk:
                    return await func(*args, **kwargs)

            return inner_coro

        return decorator



from typing import Callable, Literal

from purgatory.domain.messages.base import Event

CircuitName = str
TTL = float
Threshold = int


StateName = Literal["opened", "closed", "half-opened"]
Hook = Callable[
    [
        CircuitName,
        Literal["circuit_breaker_created", "state_changed", "failed", "recovered"],
        Event,
    ],
    None,
]



from purgatory.domain.messages.commands import CreateCircuitBreaker
from purgatory.domain.messages.events import (
    CircuitBreakerCreated,
    CircuitBreakerFailed,
    CircuitBreakerRecovered,
    ContextChanged,
)
from purgatory.domain.model import Context
from purgatory.service._sync.unit_of_work import SyncAbstractUnitOfWork


def register_circuit_breaker(
    cmd: CreateCircuitBreaker, uow: SyncAbstractUnitOfWork
) -> Context:
    """
    Register circuit breaker in the repository

    when receiving the CreateCircuitBreaker command.
    """
    ret = Context(cmd.name, cmd.threshold, cmd.ttl)
    uow.contexts.register(ret)
    uow.contexts.messages.append(
        CircuitBreakerCreated(cmd.name, cmd.threshold, cmd.ttl)
    )
    return ret


def save_circuit_breaker_state(
    evt: ContextChanged, uow: SyncAbstractUnitOfWork
) -> None:
    """
    Save the circuit breaker state in the repository

    when receiving the ContextChanged event.
    """
    uow.contexts.update_state(evt.name, evt.state, evt.opened_at)


def inc_circuit_breaker_failure(
    evt: CircuitBreakerFailed, uow: SyncAbstractUnitOfWork
) -> None:
    """
    Increment the number of failure in the repository

    when receiving the CircuitBreakerFailed event.
    """
    uow.contexts.inc_failures(evt.name, evt.failure_count)


def reset_failure(evt: CircuitBreakerRecovered, uow: SyncAbstractUnitOfWork) -> None:
    """
    Reset the number of failure in the repository

    when receiving the CircuitBreakerRecovered event.
    """
    uow.contexts.reset_failure(evt.name)



import abc
import json
from typing import Any, Optional

from purgatory.domain.messages.base import Message
from purgatory.domain.model import Context
from purgatory.service._redis import SyncRedis
from purgatory.typing import CircuitName


class ConfigurationError(RuntimeError):
    pass


class SyncAbstractRepository(abc.ABC):
    messages: list[Message]

    def initialize(self) -> None:  # noqa B027
        """Override to initialize the repository asynchronously"""

    @abc.abstractmethod
    def get(self, name: CircuitName) -> Optional[Context]:
        """Load breakers from the repository."""

    @abc.abstractmethod
    def register(self, context: Context) -> None:
        """Add a circuit breaker into the repository."""

    @abc.abstractmethod
    def update_state(
        self,
        name: str,
        state: str,
        opened_at: Optional[float],
    ) -> None:
        """Sate the new staate of the circuit breaker into the repository."""

    @abc.abstractmethod
    def inc_failures(self, name: str, failure_count: int) -> None:
        """Increment the number of failure in the repository."""

    @abc.abstractmethod
    def reset_failure(self, name: str) -> None:
        """Reset the number of failure in the repository."""


class SyncInMemoryRepository(SyncAbstractRepository):
    def __init__(self) -> None:
        self.breakers: dict[CircuitName, Context] = {}
        self.messages: list[Message] = []

    def get(self, name: CircuitName) -> Optional[Context]:
        """Add a circuit breaker into the repository."""
        return self.breakers.get(name)

    def register(self, context: Context) -> None:
        """Add a circuit breaker into the repository."""
        self.breakers[context.name] = context

    def update_state(
        self,
        name: str,
        state: str,
        opened_at: Optional[float],
    ) -> None:
        """Because the get method return the object directly, nothing to do here."""

    def inc_failures(self, name: str, failure_count: int) -> None:
        """Because the get method return the object directly, nothing to do here."""

    def reset_failure(self, name: str) -> None:
        """Reset the number of failure in the repository."""


class SyncRedisRepository(SyncAbstractRepository):
    def __init__(self, url: str) -> None:
        try:
            from redis import asyncio as aioredis
        except ImportError as exc:
            raise ConfigurationError(  # coverage: ignore
                "redis extra dependencies not installed."
            ) from exc
        self.redis: SyncRedis = aioredis.from_url(url)  # type: ignore
        self.messages = []
        self.prefix = "cbr::"

    def initialize(self) -> None:
        self.redis.initialize()  # type: ignore

    def get(self, name: CircuitName) -> Optional[Context]:
        """Add a circuit breaker into the repository."""
        data: Any = self.redis.get(f"{self.prefix}{name}")
        if not data:
            return None
        breaker = json.loads(data or "{}")
        failure_count: int = self.redis.get(f"{self.prefix}{name}::failure_count") or 0
        if failure_count:
            breaker["failure_count"] = int(failure_count)
        cbreaker = Context(**breaker)
        return cbreaker

    def register(self, context: Context) -> None:
        """Add a circuit breaker into the repository."""
        data = json.dumps(
            {
                "name": context.name,
                "threshold": context.threshold,
                "ttl": context.ttl,
                "state": context.state,
                "opened_at": context.opened_at,
            }
        )
        self.redis.set(f"{self.prefix}{context.name}", data)

    def update_state(
        self,
        name: str,
        state: str,
        opened_at: Optional[float],
    ) -> None:
        """Store the new state in the repository."""
        data: str = self.redis.get(f"{self.prefix}{name}") or "{}"
        breaker = json.loads(data)
        breaker["state"] = state
        breaker["opened_at"] = opened_at
        self.redis.set(f"{self.prefix}{name}", json.dumps(breaker))

    def inc_failures(self, name: str, failure_count: int) -> None:
        """Store the new state in the repository."""
        self.redis.incr(f"{self.prefix}{name}::failure_count")

    def reset_failure(self, name: str) -> None:
        """Reset the number of failure in the repository."""
        self.redis.set(f"{self.prefix}{name}::failure_count", "0")



"""Unit of work"""

from __future__ import annotations

import abc
from collections.abc import Generator
from types import TracebackType
from typing import Optional

from purgatory.domain.messages import Message
from purgatory.service._async.repository import (
    AsyncAbstractRepository,
    AsyncInMemoryRepository,
    AsyncRedisRepository,
)


class AsyncAbstractUnitOfWork(abc.ABC):
    contexts: AsyncAbstractRepository

    def collect_new_events(self) -> Generator[Message, None, None]:
        while self.contexts.messages:
            yield self.contexts.messages.pop(0)

    async def initialize(self) -> None:  # noqa B027
        """Override to initialize  repositories."""

    async def __aenter__(self) -> AsyncAbstractUnitOfWork:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        """Rollback in case of exception."""
        if exc:
            await self.rollback()

    @abc.abstractmethod
    async def commit(self) -> None:
        """Commit the transation."""

    @abc.abstractmethod
    async def rollback(self) -> None:
        """Rollback the transation."""


class AsyncInMemoryUnitOfWork(AsyncAbstractUnitOfWork):
    def __init__(self) -> None:
        self.contexts = AsyncInMemoryRepository()

    async def commit(self) -> None:
        """Do nothing."""

    async def rollback(self) -> None:
        """Do nothing."""


class AsyncRedisUnitOfWork(AsyncAbstractUnitOfWork):
    def __init__(self, url: str) -> None:
        self.contexts = AsyncRedisRepository(url)

    async def initialize(self) -> None:
        await self.contexts.initialize()

    async def commit(self) -> None:
        """Do nothing."""

    async def rollback(self) -> None:
        """Do nothing."""



"""
Circuit breaker state model.

The state model is implemented using the State pattern from the Gang Of Four.
"""

import abc
import time
from dataclasses import dataclass
from types import TracebackType
from typing import Callable, Optional, Union, cast

from purgatory.domain.messages.base import Event
from purgatory.domain.messages.events import (
    CircuitBreakerFailed,
    CircuitBreakerRecovered,
    ContextChanged,
)
from purgatory.typing import TTL, CircuitName, StateName, Threshold

ExcludeExcType = type[BaseException]
ExcludeTypeFunc = tuple[ExcludeExcType, Callable[..., bool]]
ExcludeType = list[
    Union[
        ExcludeExcType,
        ExcludeTypeFunc,
    ]
]

CLOSED: StateName = "closed"
OPENED: StateName = "opened"
HALF_OPENED: StateName = "half-opened"


class Context:
    name: CircuitName
    threshold: Threshold
    ttl: TTL
    messages: list[Event]
    exclude_list: ExcludeType

    def __init__(
        self,
        name: CircuitName,
        threshold: Threshold,
        ttl: TTL,
        state: StateName = "closed",
        failure_count: int = 0,
        opened_at: Optional[float] = None,
        exclude: Optional[ExcludeType] = None,
    ) -> None:
        self.name = name
        self.ttl = ttl
        self.threshold = threshold

        self._state: State
        if state == OpenedState.name:
            self._state = OpenedState(name)
        elif state == HalfOpenedState.name:
            self._state = HalfOpenedState()
        else:
            self._state = ClosedState()

        if opened_at is not None:
            self._state.opened_at = opened_at
        self._state.failure_count = failure_count
        self.messages = []
        self.exclude_list = exclude or []

    @property
    def state(self) -> StateName:
        return cast(StateName, self._state.name)

    @property
    def opened_at(self) -> Optional[float]:
        return self._state.opened_at

    @property
    def failure_count(self) -> Optional[int]:
        return self._state.failure_count

    def set_state(self, state: "State") -> None:
        self._state = state
        self.messages.append(
            ContextChanged(
                self.name,
                self.state,
                state.opened_at,
            )
        )

    def mark_failure(self, failure_count: int) -> None:
        self.messages.append(
            CircuitBreakerFailed(
                self.name,
                failure_count,
            )
        )

    def recover_failure(self) -> None:
        self.messages.append(
            CircuitBreakerRecovered(
                self.name,
            )
        )

    def handle_new_request(self) -> None:
        self._state.handle_new_request(self)

    def handle_exception(self, exc: BaseException) -> None:
        failed = True
        for exctype_func in self.exclude_list:
            if isinstance(exctype_func, tuple):
                exctype, func = cast(ExcludeTypeFunc, exctype_func)
            else:
                exctype, func = (
                    cast(ExcludeExcType, exctype_func),
                    cast(
                        Callable[[BaseException], bool],
                        lambda exc: True,  # type: ignore
                    ),
                )

            if isinstance(exc, exctype):
                failed = not func(exc)
                break
        if failed:
            self._state.handle_exception(self, exc)
        else:
            self._state.handle_end_request(self)

    def handle_end_request(self) -> None:
        self._state.handle_end_request(self)

    def __enter__(self) -> "Context":
        self.handle_new_request()
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        if exc:
            return self.handle_exception(exc)
        else:
            return self.handle_end_request()

    def __repr__(self) -> str:
        return (
            f"<Context "
            f'name="{self.name}" '
            f'state="{self.state}" '
            f'threshold="{self.threshold}" ttl="{self.ttl}">'
        )

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Context)
            and self.name == other.name
            and self.threshold == other.threshold
            and self.ttl == other.ttl
        )


class State(abc.ABC):
    failure_count: Optional[int] = None
    opened_at: Optional[float] = None
    name: str = ""

    @abc.abstractmethod
    def handle_new_request(self, context: Context) -> None:
        """Handle new code block"""

    @abc.abstractmethod
    def handle_exception(self, context: Context, exc: BaseException) -> None:
        """Handle exception during the code block"""

    @abc.abstractmethod
    def handle_end_request(self, context: Context) -> None:
        """Handle proper execution after the code block"""


@dataclass
class ClosedState(State):
    """In closed state, track for failure."""

    failure_count: int
    name: StateName = CLOSED

    def __init__(self) -> None:
        self.failure_count = 0

    def handle_new_request(self, context: Context) -> None:
        """When the circuit is closed, the new request has no incidence"""

    def handle_exception(self, context: Context, exc: BaseException) -> None:
        self.failure_count += 1
        context.mark_failure(self.failure_count)
        if self.failure_count >= context.threshold:
            opened = OpenedState(context.name)
            context.set_state(opened)

    def handle_end_request(self, context: Context) -> None:
        """Reset in case the request is ok"""
        if self.failure_count > 0:
            context.recover_failure()
        self.failure_count = 0


@dataclass
class OpenedState(State, Exception):
    """In open state, reopen after a TTL."""

    opened_at: float
    name: StateName = OPENED

    def __init__(self, circuit_name: CircuitName) -> None:
        Exception.__init__(self, f"Circuit {circuit_name} is open")
        self.opened_at = time.time()
        self.circuit_name = circuit_name

    def handle_new_request(self, context: Context) -> None:
        closed_at = self.opened_at + context.ttl
        if time.time() > closed_at:
            context.set_state(HalfOpenedState())
            return context.handle_new_request()
        raise self

    def handle_exception(self, context: Context, exc: BaseException) -> None:
        """
        When the circuit is opened, the OpenState is raised before entering.

        This function is never called.
        """

    def handle_end_request(self, context: Context) -> None:
        """
        When the circuit is opened, the OpenState is raised before entering.

        This function is never called.
        """


@dataclass
class HalfOpenedState(State):
    """In half open state, decide to reopen or to close."""

    name: StateName = HALF_OPENED

    def handle_new_request(self, context: Context) -> None:
        """In half open state, we reset the failure counter to restart 0."""

    def handle_exception(self, context: Context, exc: BaseException) -> None:
        """If an exception happens, then the circuit is reopen directly."""
        opened = OpenedState(context.name)
        context.set_state(opened)

    def handle_end_request(self, context: Context) -> None:
        """Otherwise, the circuit is closed, back to normal."""
        context.recover_failure()
        context.set_state(ClosedState())
